@@ -127,6 +127,7 @@ typedef struct _GridScreen
     XRectangle     currentRect;
     GridProps      props;
     CompWindow *w;
+    int lastOutput;
     Bool drawing;
     Animation anim;
     Bool animating;
@@ -160,18 +161,17 @@ constrainSize (CompWindow *w,
 	       XRectangle *slot,
 	       XRectangle *rect)
 {
-    XRectangle workarea;
+    GRID_SCREEN (w->screen);
     XRectangle r;
     int        cw, ch;
 
-    getWorkareaForOutput (w->screen, outputDeviceForWindow (w), &workarea);
     slotToRect (w, slot, &r);
 
     if (constrainNewWindowSize (w, r.width, r.height, &cw, &ch))
     {
 	/* constrained size may put window offscreen, adjust for that case */
-	int dx = r.x + cw - workarea.width - workarea.x + w->input.right;
-	int dy = r.y + ch - workarea.height - workarea.y + w->input.bottom;
+	int dx = r.x + cw - gs->workarea.width - gs->workarea.x + w->input.right;
+	int dy = r.y + ch - gs->workarea.height - gs->workarea.y + w->input.bottom;
 
 	if ( dx > 0 )
 	    r.x -= dx;
@@ -188,7 +188,8 @@ constrainSize (CompWindow *w,
 
 static void
 getTargetRect (CompWindow *cw,
-	       GridType	  where)
+	       GridType	  where,
+	       Bool setWorkarea)
 {
     GRID_SCREEN (cw->screen);
 
@@ -197,8 +198,9 @@ getTargetRect (CompWindow *cw,
     DEBUG_PRINT ((gridOut, "\nPressed KP_%i\n", where));
 
     /* get current available area */
-    getWorkareaForOutput (cw->screen, outputDeviceForWindow(cw), &gs->workarea);
-    DEBUG_RECT (workarea);
+    if (setWorkarea)
+		getWorkareaForOutput (cw->screen, outputDeviceForWindow(cw), &gs->workarea);
+    DEBUG_RECT (gs->workarea);
 
     /* Convention:
      * xxxSlot include decorations (it's the screen area occupied)
@@ -213,7 +215,7 @@ getTargetRect (CompWindow *cw,
     gs->desiredSlot.x = gs->workarea.x + gs->props.gravityRight *
 	                (gs->workarea.width / gs->props.numCellsX);
     gs->desiredSlot.width = gs->workarea.width / gs->props.numCellsX;
-    DEBUG_RECT (desiredSlot);
+    DEBUG_RECT (gs->desiredSlot);
 
     /* Adjust for constraints and decorations */
     constrainSize (cw, &gs->desiredSlot, &gs->desiredRect);
@@ -247,7 +249,8 @@ static void sendMaximizationRequest (CompWindow *w)
 
 static void
 gridCommonWindow (CompWindow *cw,
-                  GridType   where)
+                  GridType   where,
+                  Bool setWorkarea)
 {
     GRID_SCREEN (cw->screen);
 
@@ -264,7 +267,7 @@ gridCommonWindow (CompWindow *cw,
 	    unsigned int valueMask = 0;
 	    int desiredState = 0;
 
-	    getTargetRect (cw, where);
+	    getTargetRect (cw, where, setWorkarea);
 
 	    XWindowChanges xwc;
 
@@ -275,11 +278,11 @@ gridCommonWindow (CompWindow *cw,
 	    gs->currentRect.y = cw->serverY;
 	    gs->currentRect.width  = cw->serverWidth;
 	    gs->currentRect.height = cw->serverHeight;
-	    DEBUG_RECT (currentRect);
+	    DEBUG_RECT (gs->currentRect);
 
-	    if ((gs->desiredRect.y == gs->currentRect.y &&
-		gs->desiredRect.height == gs->currentRect.height) &&
-		gridGetCycleSizes(cw->screen->display))
+	    if ((gs->desiredRect.y == gs->currentRect.y + cw->clientFrame.top &&
+			gs->desiredRect.height == gs->currentRect.height - (cw->clientFrame.top + cw->clientFrame.bottom)) &&
+			gridGetCycleSizes(cw->screen->display))
 	    {
 		int slotWidth33 = gs->workarea.width / 3;
 		int slotWidth66 = gs->workarea.width - slotWidth33;
@@ -288,8 +291,9 @@ gridCommonWindow (CompWindow *cw,
 
 		if (gs->props.numCellsX == 2) /* keys (1, 4, 7, 3, 6, 9) */
 		{
-		    if (gs->currentRect.width == gs->desiredRect.width &&
-			gs->currentRect.x == gs->desiredRect.x)
+		    if (gs->currentRect.width == gs->desiredRect.width +
+		        cw->clientFrame.left + cw->clientFrame.right &&
+			    gs->currentRect.x == gs->desiredRect.x - cw->clientFrame.left)
 		    {
 			gs->desiredSlot.width = slotWidth66;
 			gs->desiredSlot.x = gs->workarea.x +
@@ -318,8 +322,9 @@ gridCommonWindow (CompWindow *cw,
 			DEBUG_RECT (slot66);
 			DEBUG_RECT (rect66);
 
-			if (gs->currentRect.width == rect66.width &&
-			    gs->currentRect.x == rect66.x)
+			if (gs->currentRect.width == (rect66.width +
+			    cw->clientFrame.left + cw->clientFrame.right) &&
+			    gs->currentRect.x == rect66.x - cw->clientFrame.left)
 			{
 			    gs->desiredSlot.width = slotWidth33;
 			    gs->desiredSlot.x = gs->workarea.x +
@@ -329,42 +334,51 @@ gridCommonWindow (CompWindow *cw,
 		}
 		else /* keys (2, 5, 8) */
 		{
-		    if (gs->currentRect.width == gs->desiredRect.width &&
-			gs->currentRect.x == gs->desiredRect.x)
+		    if (gs->currentRect.width == (gs->desiredRect.width +
+			    cw->clientFrame.left + cw->clientFrame.right) &&
+			    gs->currentRect.x == gs->desiredRect.x - cw->clientFrame.left)
 		    {
 			gs->desiredSlot.width = slotWidth33;
 			gs->desiredSlot.x = gs->workarea.x + slotWidth33;
 		    }
 		}
 		constrainSize (cw, &gs->desiredSlot, &gs->desiredRect);
-		DEBUG_RECT (gs->desiredRect);
+	        DEBUG_RECT (gs->desiredRect);
 	    }
 
-	    xwc.x = gs->desiredRect.x;
-	    xwc.y = gs->desiredRect.y;
-	    xwc.width  = gs->desiredRect.width;
-	    xwc.height = gs->desiredRect.height;
-
-	    if (cw->mapNum)
-		sendSyncRequest (cw);
+	    /*Note that ClientFrame extents must also be computed in the "if"
+	     *blocks for cycling through window sizes or the checks fail
+	     */
+	    xwc.x = gs->desiredRect.x - cw->clientFrame.left;
+	    xwc.y = gs->desiredRect.y - cw->clientFrame.top;
+	    xwc.width  = gs->desiredRect.width + cw->clientFrame.left + cw->clientFrame.right;
+	    xwc.height = gs->desiredRect.height + cw->clientFrame.top + cw->clientFrame.bottom;
 
 	    if (where == GridRight || where == GridLeft)
 	    {
 		desiredState = CompWindowStateMaximizedVertMask;
-		valueMask = CWX | CWWidth;
+		valueMask = CWX;
 	    }
-	    else if (where == GridTop || where == GridBottom)
+	    else if (where == GridTop || where == GridBottom) 
 	    {
 		desiredState = CompWindowStateMaximizedHorzMask;
-		valueMask = CWY | CWHeight;
+		valueMask = CWY;
 	    }
 	    else
 	    {
 		desiredState = 0;
-		valueMask = CWX | CWY | CWWidth | CWHeight;
-	    }
+		valueMask = CWX | CWY;
+            }
 
-	   if (cw->state != desiredState)
+            if (xwc.width != cw->serverWidth)
+                valueMask |= CWWidth;
+            if (xwc.height != cw->serverHeight)
+                valueMask |= CWHeight;
+
+            if (cw->mapNum && (valueMask & (CWWidth | CWHeight))) 
+                sendSyncRequest (cw);
+
+	    if (cw->state != desiredState)
 		maximizeWindow (cw, desiredState);
 
 	    /* TODO: animate move+resize */
@@ -388,7 +402,7 @@ gridCommon (CompDisplay	    *d,
     xid = getIntOptionNamed (option, nOption, "window", 0);
     cw  = findWindowAtDisplay (d, xid);
 
-    gridCommonWindow(cw, where);
+    gridCommonWindow(cw, where, TRUE);
 
     return TRUE;
 }
@@ -441,92 +455,109 @@ gridHandleEvent (CompDisplay *d,
     GridType     where;
     GRID_DISPLAY (d);
 
-    if (event->type == MotionNotify)
-    {
+    if (event->type != MotionNotify)
+		goto out;
+
 	CompScreen *s;
 	s = findScreenAtDisplay (d, event->xmotion.root);
-	if (s)
+
+	if (!s)
+		goto out;
+
+    GRID_SCREEN (s);
+
+    if (!gs->grabIsMove)
+		goto out;
+
+	int o = gridGetOutputSelectMousePointer(d) ?
+			outputDeviceForPoint (s, pointerX, pointerY) :
+			outputDeviceForWindow (gs->w);
+	BOX extents = s->outputDev[o].region.extents;
+	if (gridGetOutputSelectMousePointer(d))
+		getWorkareaForOutput (s, o, &gs->workarea);
+
+	Bool top = (pointerY < extents.y1 + gridGetTopEdgeThreshold(d) &&
+				pointerY > extents.y1 - gridGetBottomEdgeThreshold(d));
+	Bool bottom = (pointerY > (extents.y2 - gridGetBottomEdgeThreshold(d)) &&
+					pointerY < (extents.y2 + gridGetTopEdgeThreshold(d)));
+	Bool left = (pointerX < extents.x1 + gridGetLeftEdgeThreshold(d) &&
+				pointerX > extents.x1 - gridGetRightEdgeThreshold(d));
+	Bool right = (pointerX > (extents.x2 - gridGetRightEdgeThreshold(d)) &&
+				 pointerX < (extents.x2 + gridGetLeftEdgeThreshold(d)));
+
+	/* detect corners first */
+	if (top)
 	{
-	    GRID_SCREEN (s);
-	    if (gs->grabIsMove)
-	    {
-		/* detect corners first */
-		/* Bottom Left */
-		if (pointerY > (s->height - gridGetBottomEdgeThreshold(d)) &&
-		    pointerX < gridGetLeftEdgeThreshold(d))
-		    gs->edge = BottomLeft;
-		/* Bottom Right */
-		else if (pointerY > (s->height - gridGetBottomEdgeThreshold(d)) &&
-			 pointerX > (s->width - gridGetRightEdgeThreshold(d)))
-		    gs->edge = BottomRight;
-		/* Top Left */
-		else if (pointerY < gridGetTopEdgeThreshold(d) && pointerX < gridGetLeftEdgeThreshold(d))
-		    gs->edge = TopLeft;
-		/* Top Right */
-		else if (pointerY < gridGetTopEdgeThreshold(d) &&
-			 pointerX > (s->width - gridGetRightEdgeThreshold(d)))
-		    gs->edge = TopRight;
-		/* Left */
-		else if (pointerX < gridGetLeftEdgeThreshold(d))
-		    gs->edge = Left;
-		/* Right */
-		else if (pointerX > (s->width - gridGetRightEdgeThreshold(d)))
-		    gs->edge = Right;
-		/* Top */
-		else if (pointerY < gridGetTopEdgeThreshold(d))
-		    gs->edge = Top;
-		/* Bottom */
-		else if (pointerY > (s->height - gridGetBottomEdgeThreshold(d)))
-		    gs->edge = Bottom;
-		/* No Edge */
+		if (left)
+			gs->edge = TopLeft;
+		else if (right)
+			gs->edge = TopRight;
 		else
-		    gs->edge = NoEdge;
+			gs->edge = Top;
+	} else
+	if (bottom)
+	{
+		if (left)
+			gs->edge = BottomLeft;
+		else if (right)
+			gs->edge = BottomRight;
+		else
+			gs->edge = Bottom;
+	} else
+	if (left)
+		gs->edge = Left;
+	else if (right)
+		gs->edge = Right;
+	else
+		gs->edge = NoEdge;
 
-		/* detect edge region change */
-		if (gs->lastEdge != gs->edge)
+	/* detect edge region change */
+	if (gs->lastEdge != gs->edge || o != gs->lastOutput)
+	{
+		if (gs->edge != NoEdge)
 		{
-		    if (gs->edge != NoEdge)
-		    {
-			where = edgeToGridType(d, gs->edge);
+		where = edgeToGridType(d, gs->edge);
 
-			/* treat Maximize visual indicator same as GridCenter */
-			if (where == GridMaximize)
-			    where=GridCenter;
+		/* treat Maximize visual indicator same as GridCenter */
+		if (where == GridMaximize)
+			where=GridCenter;
 
-			getTargetRect (gs->w, where);
+		/* Do not show animation for GridUnknown (no action) */
+		if (where == GridUnknown)
+			return;
 
-			gs->anim.duration = gridGetAnimationDuration (d);
-			gs->anim.timer = gs->anim.duration;
-			gs->anim.opacity = 0.0f;
-			gs->anim.progress = 0.0f;
-			gs->anim.currentRect.x = gs->w->serverX;
-			gs->anim.currentRect.y = gs->w->serverY;
-			gs->anim.currentRect.width = gs->w->serverWidth;
-			gs->anim.currentRect.height = gs->w->serverHeight;
-			gs->anim.targetRect = gs->desiredSlot;
-			gs->anim.fromRect.x = gs->w->serverX - gs->w->input.left;
-			gs->anim.fromRect.y = gs->w->serverY - gs->w->input.top;
-			gs->anim.fromRect.width = gs->w->serverWidth +
-						  gs->w->input.left +
-						  gs->w->input.right +
-						  gs->w->serverBorderWidth * 2;
-			gs->anim.fromRect.height = gs->w->serverHeight +
-						   gs->w->input.top +
-						   gs->w->input.bottom +
-						   gs->w->serverBorderWidth * 2;
-			gs->animating = TRUE;
-			gs->anim.fadingOut = FALSE;
-		    }
-		    else
-			gs->anim.fadingOut = TRUE;
+		getTargetRect (gs->w, where, !gridGetOutputSelectMousePointer(d));
 
-		    gs->lastEdge = gs->edge;
+		gs->anim.duration = gridGetAnimationDuration (d);
+		gs->anim.timer = gs->anim.duration;
+		gs->anim.opacity = 0.0f;
+		gs->anim.progress = 0.0f;
+		gs->anim.currentRect.x = gs->w->serverX;
+		gs->anim.currentRect.y = gs->w->serverY;
+		gs->anim.currentRect.width = gs->w->serverWidth;
+		gs->anim.currentRect.height = gs->w->serverHeight;
+		gs->anim.targetRect = gs->desiredSlot;
+		gs->anim.fromRect.x = gs->w->serverX - gs->w->input.left;
+		gs->anim.fromRect.y = gs->w->serverY - gs->w->input.top;
+		gs->anim.fromRect.width = gs->w->serverWidth +
+					  gs->w->input.left +
+					  gs->w->input.right +
+					  gs->w->serverBorderWidth * 2;
+		gs->anim.fromRect.height = gs->w->serverHeight +
+					   gs->w->input.top +
+					   gs->w->input.bottom +
+					   gs->w->serverBorderWidth * 2;
+		gs->animating = TRUE;
+		gs->anim.fadingOut = FALSE;
 		}
+		else
+		gs->anim.fadingOut = TRUE;
 
-	    }
+		gs->lastEdge = gs->edge;
 	}
-    }
+	gs->lastOutput = o;
 
+out:
     UNWRAP (gd, d, handleEvent);
     (*d->handleEvent) (d, event);
     WRAP (gd, d, handleEvent, gridHandleEvent);
@@ -547,6 +578,7 @@ gridWindowGrabNotify (CompWindow   *w,
     {
 	gs->grabIsMove = TRUE;
 	gs->w = w;
+	gs->lastOutput = outputDeviceForWindow (w);
     }
 
     UNWRAP (gs, s, windowGrabNotify);
@@ -568,7 +600,7 @@ gridWindowUngrabNotify (CompWindow *w)
 
 	if (gs->edge != NoEdge)
 	{
-	    gridCommonWindow (w, edgeToGridType(d, gs->edge));
+	    gridCommonWindow (w, edgeToGridType(d, gs->edge), FALSE);
 	    gs->anim.fadingOut = TRUE;
 	}
     }
